# Copyright 2025 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class QubitOutOfRangeError(Exception):
    """Raised when a qubit index is outside the defined circuit range."""


class GateHasNoMatrixError(Exception):
    """Raised when a gate lacks an associated matrix representation."""


class GateNotParameterizedError(Exception):
    """Raised when attempting to set parameters on a non-parameterized gate."""


class ParametersNotEqualError(Exception):
    """Raised when parameterized gates receive mismatched parameter values."""


class InvalidParameterNameError(Exception):
    """Raised when an unknown parameter name is supplied to a gate."""


class UnsupportedGateError(Exception):
    """Raised when a gate is not supported by the target backend."""
